/******************************************************************************
 *  GTERM - GGLABS TERMINAL                                                   *
 *  A fast 80 columns terminal for the C64                                    *
 *  Copyright 2019 Gabriele Gorla                                             *
 *                                                                            *
 *  This program is free software: you can redistribute it and/or modify      *
 *  it under the terms of the GNU General Public License as published by      *
 *  the Free Software Foundation, either version 3 of the License, or         *
 *  (at your option) any later version.                                       *
 *                                                                            *
 *  GMT is distributed in the hope that it will be useful,                    *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
 *  GNU General Public License for more details.                              *
 *                                                                            *
 *  You should have received a copy of the GNU General Public License         *
 *  along with GTERM.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 ******************************************************************************/
#include <stdio.h>
#include <conio.h>
#include <c64.h>

#include "gterm.h"
#include "asc_ctype.h"

#define VERSION "0.12f"

#define SPR_BASE   (COLOR_BASE+0x3f8)

#define ACIA_DATA    ACIA_BASE.data
#define ACIA_STATUS  ACIA_BASE.status
#define ACIA_COMMAND ACIA_BASE.command
#define ACIA_CONTROL ACIA_BASE.control
#define ACIA_EXT     ACIA_BASE.ext

unsigned char turbo232, baud_idx_max;

unsigned char esc_mode;  // vt100 state machine

unsigned char priv_mode[2];  // DEC private state
unsigned char attr;   // char attribute

unsigned char g_set;   // active char set (G0/G1)
unsigned char G[2];    // G0 and G1 charset
unsigned char AM[2];   // active top and bottom margin
unsigned char tab[10]; // tab stops (80 bits)

// tx circular buffer
unsigned char send_buf[32];
unsigned char send_get, send_put;

unsigned char baud_index; // current baud rate

// storage for save/restore cursor
unsigned char col_save, row_save, attr_save;

unsigned char * const line[] = {
		       //			SCR_BASE+0,
			SCR_BASE+320*1,
			SCR_BASE+320*2,
			SCR_BASE+320*3,
			SCR_BASE+320*4,
			SCR_BASE+320*5,
			SCR_BASE+320*6,
			SCR_BASE+320*7,
			SCR_BASE+320*8,
			SCR_BASE+320*9,
			SCR_BASE+320*10,
			SCR_BASE+320*11,
			SCR_BASE+320*12,
			SCR_BASE+320*13,
			SCR_BASE+320*14,
			SCR_BASE+320*15,
			SCR_BASE+320*16,
			SCR_BASE+320*17,
			SCR_BASE+320*18,
			SCR_BASE+320*19,
			SCR_BASE+320*20,
			SCR_BASE+320*21,
			SCR_BASE+320*22,
			SCR_BASE+320*23,
			SCR_BASE+320*24,
};


const struct baud_table_t baud_table[] =
  {
   {   24, 0x18, 0x00 },
   {   96, 0x1C, 0x00 },
   {  192, 0x1E, 0x00 },
   {  384, 0x1F, 0x00 },
   {  576, 0x10, 0x42 },
   { 1152, 0x10, 0xC1 },
   //   { 2304, 0x10, 0xC0 },
  };

#define BAUD_TABLE_SIZE (sizeof(baud_table)/sizeof(struct baud_table_t))

void (*Scroll1)(void);
void (*InvScroll1)(void);

unsigned char __fastcall__ recv_fifo_fill(void)
{
  return (put-get)&0xff;
}

unsigned char __fastcall__ send_fifo_space(void)
{
  return 0x1f-(send_put-send_get)&0x1f;
}

void GenScroll1()
{
  for(ctemp1=AM[0];ctemp1<AM[1];ctemp1++) {
    line_addr=line[ctemp1];
    SRC_LINE=line[ctemp1+1];
    CopyLine();
  }
  line_addr=line[ctemp1];
  ClrLine();
  line_addr=line[ctemp1];
}

void GenInvScroll1()
{
  for(ctemp1=AM[1];ctemp1>AM[0];ctemp1--) {
    line_addr=line[ctemp1];
    SRC_LINE=line[ctemp1-1];
    CopyLine();
  }
  line_addr=line[ctemp1];
  ClrLine();
  line_addr=line[ctemp1];
}


void set_pos()
{
  if(COL>79) {
    if(priv_mode[0]&PRIV_DECAWM) {
      ++ROW;
      COL=0;
    } else {
      COL=79;
    }
  }
  if(ROW<AM[0]) ROW=AM[0];
  else if(ROW>AM[1]) {
    ROW=AM[1];
    Scroll1();
  }
  line_addr=POS(ROW,COL);
}

void set_cursor()
{
  itemp=(COL<<2)+24;
  VIC.spr0_x   = ctemp1;
  VIC.spr_hi_x = ctemp2;
  VIC.spr0_y   = (ROW<<3)+64;
}

void ser_set()
{
  ACIA_CONTROL=baud_table[baud_index].ctrl;
  ACIA_EXT=baud_table[baud_index].ext;
  
  if(baud_table[baud_index].ext & BT_EXT_POLL_MASK) {
    // no parity, no echo, no tx irq, no rts, no rx irq, en dtr
    ACIA_COMMAND=0x03;
  } else {
    // no parity, no echo, no tx irq, en rts, en rx irq, en dtr
    ACIA_COMMAND=0x09;
  }
}

void ser_init()
{
  unsigned char j;
  
  get=0;
  put=0;
  send_get=0;
  send_put=0;
  
  ACIA_CONTROL=0;
  
  if(ACIA_EXT) {
    turbo232=1;
    baud_idx_max=BAUD_TABLE_SIZE-1;
  } else {
    turbo232=0;
    for(j=BAUD_TABLE_SIZE-1;j>0;j--)
      if(!(baud_table[j].ext&BT_EXT_T232_REQ)) break;
    baud_idx_max=j;
  }
  
  baud_index=3;
  ser_set();
}

void ser_send()
{
  unsigned char st;
  st=ACIA_STATUS;
  if(st&0x10) {
    ACIA_DATA=send_buf[send_get];
    send_get++;
    send_get&=0x1F;
  }
}

void send_str(unsigned char *str)
{
  // fixme: check for buffer full
  while(*str) {
    send_buf[send_put]=conv_table[*str];
    str++;
    send_put++;
    send_put&=0x1F;
  }
}

void ser_put_char(unsigned char c)
{
  // fixme: check for buffer full
  send_buf[send_put]=conv_table[c];
  send_put++;
  send_put&=0x1F;
}

void send_cursor(unsigned char c)
{
  if(send_fifo_space()<6) return;

  ser_put_char(0x1B);
  
  if(priv_mode[0]&0x01) // DECCKM cursor key mode
    ser_put_char('O');
  else
    ser_put_char('[');

  ser_put_char(c);
}

void send_func(unsigned char c)
{
  if(send_fifo_space()<6) return;
  ser_put_char(0x1B);
  ser_put_char('O');
  ser_put_char(c);
}

void conv_str(unsigned char *str)
{
  while(*str) {
    *str=conv_table[*str];
    str++;
  }
}


unsigned char * line_save;
unsigned char loc_csave;

void PutTitleAt(unsigned char c)
{
  conv_str(STR_BUF);
  loc_csave=COL;
  line_save=line_addr;
  COL=c;
  line_addr=SCR_BASE+((c>>1)<<3);
  PutLine();
  line_addr=line_save;
  COL=loc_csave;
}

void PutLineAt(unsigned char r, unsigned char c)
{
  conv_str(STR_BUF);
  loc_csave=COL;
  line_save=line_addr;
  COL=c;
  line_addr=POS(r,COL);
  PutLine();
  line_addr=line_save;
  COL=loc_csave;
}

void lf()
{
  ++ROW;
  set_cursor();
}

unsigned char * line_temp;
unsigned char ctemp;

void title()
{
  sprintf(STR_BUF,"GGLABS GTERM V"VERSION" - %4d00 8N1  ",baud_table[baud_index].baud);
  PutTitleAt(0);
}


void Bitmap_En()
{
  VIC.ctrl1=0x3B; // enable bitmap, no extended color, no blank, 25 rows, ypos=3
  //  CIA2.pra=0x00;  // selects VIC page 0xC000-0xFFFF
  CIA2.pra=0x2;  // selects VIC page 0x4000-0x7FFF
  
  //  VIC.addr=0x68;  
  VIC.addr=0x78;  // color ram at base +0x1c00, bitmap at base + 0x2000
  VIC.ctrl2=0xC8; // no multicolor, 40 cols, xpos=0
}

void Bitmap_Dis()
{
  VIC.ctrl1=0x1B; // disable bitmap, no extended color, no blank, 25 rows, ypos=3
  CIA2.pra=0x03;  // selects vic page 0x0000-0x3fff
  VIC.addr=0x15;  // screen base at 0x0400, char def at $0x1400
}

void Bitmap_Setup()
{
  //  unsigned int i;
  //  unsigned char x;
  
  line_addr=SCR_BASE;
  ClrLine();
  ClrScr();

  // clear color RAM
  c1=0x01;
  SetColor();

  // white on red title
  //  for(ctemp1=0;ctemp1<40;ctemp1++)
  //    COLOR_BASE[ctemp1]=0x12;
  memset8(COLOR_BASE,0x12,40);

  // cursor sprite
  //   for(ctemp1=0;ctemp1<64;ctemp1++)
  //     SPR_DATA[ctemp1]=0;
  memset8(SPR_DATA,0,64);
  
  SPR_DATA[0]=0xf0;

  VIC.spr0_color=0x00;
  SPR_BASE[0]=111;  
}

const void (* const esc_func[7])(void) =
  {
   NULL,
   escape_mode_imm,
   escape_mode_csi,
   escape_mode_hash,
   escape_mode_priv,
   escape_mode_scs,
   escape_mode_scs
  };

int main()
{
  unsigned char c;

  // setup the NMI vector
  *((unsigned int *)0x0318)=(unsigned int)NMI;

#ifdef DEBUG
  printf("GTERM starting\n");
  printf("bitmap at 0x%04X\n",SCR_BASE);
  printf("charset at 0x%04X\n",soft80_charset);
  printf("conv_table at 0x%04X\n",conv_table);
  printf("isxx table at 0x%04X\n",isxx);

  cgetc();
#endif
  
  Bitmap_Setup();
  Bitmap_En();
  
  ser_init();
  title();
   
  esc_mode=0;
  COL=0;
  ROW=0;
  attr=0;
  col_save=0;
  row_save=0;
  attr_save=0;
  priv_mode[0]=PRIV_DECAWM;
  priv_mode[1]=0;
  g_set=0;
  G[0]=CHAR_US;
  G[1]=CHAR_US;
  AM[0]=0;
  AM[1]=23;

  // default tab is every 8 columns
  memset8(tab,1,10);
  tab[0]=0;

  Scroll1=FS_Scroll1;
  InvScroll1=FS_InvScroll1;
  
  sprintf(STR_BUF,"GGLABS Terminal V%s - http://gglabs.us",VERSION);
  PutLineAt(0,0);
  sprintf(STR_BUF,"A modern VT100 terminal emulator for the Commodore 64");
  PutLineAt(1,0);
  sprintf(STR_BUF,"Copyright (c) 2019 Gabriele Gorla");
  PutLineAt(2,0);
  
  sprintf(STR_BUF,"This program is free software: you can redistribute it and/or modify it under");
  PutLineAt(4,0);
  sprintf(STR_BUF,"the terms of the GNU General Public License as published by the Free Software");
  PutLineAt(5,0);
  sprintf(STR_BUF,"Foundation, either License version 3, or (at your option) any later version.");
  PutLineAt(6,0);

  sprintf(STR_BUF,"ACIA Cartridge: %s",(turbo232)?"Turbo232":"Swiftlink");
  PutLineAt(8,0);

  ROW=10;
  set_cursor();
  VIC.spr_ena=0x01;


  for(;;) { // main loop
    
    //  if(put!=get) {
    if(recv_fifo_fill()) {
      if(!esc_mode) {
#if 1
	// process printable character in pairs
	//	if(((COL&1)==0) && (fifo_fill()>1)) {
	if((COL&1)==0) {
	  while(recv_fifo_fill()>1) {
	    c1=SER_BUF[get];
	    c2=SER_BUF[(get+1)&0xff];
	    if(isprint(c1)&&isprint(c2)) {
	      if(G[g_set]==CHAR_GRP) {
		c1-=isdecgfx(c1)?0x5F:0;
		c2-=isdecgfx(c2)?0x5F:0;
	      }
	      set_pos();
	      //	      attr|=ATTR_INVERSE;
	      Put2Char80();
	      //	      attr&=~ATTR_INVERSE;
	      get+=2;
	    } else break;
	  }
	}
#endif
	// odd position character, last character or control character
	if(recv_fifo_fill()) { 
	  c1=SER_BUF[get];
	  ++get;
	  if(isprint(c1)) {
	    if(G[g_set]==CHAR_GRP) {
	      c1-=isdecgfx(c1)?0x5F:0;
	    }
	    set_pos();
	    PutChar80();	    
	  } else {
	    parse_control();
	  }
	}
      } else {
	c1=SER_BUF[get];
	++get;
	(*esc_func[esc_mode])();
      }
      set_cursor(); // should be able to optimize
    } else {
      // input buffer empty
      if(baud_table[baud_index].ext & BT_EXT_POLL_MASK) {
	ser_recv();
      } else {
	// release RTS
	ACIA_COMMAND=0x09;
	*(char *)(SCR_BASE-0x400+38)=1;
      }
    } 
    if(kbhit()) {
      c=cgetc();
      
#ifdef DEBUG
      sprintf(STR_BUF,"%02x  %02x %02x ",c,priv_mode[0],priv_mode[1]);
      PutTitleAt(36);
#endif
      
      switch(c) {
      case 0x11: // down
	send_cursor('B');
	break;
      case 0x1D: // right
	send_cursor('C');
	break;
      case 0x91: // up
	send_cursor('A');
	break;
      case 0x9D: // left
	send_cursor('D');
	break;
	
      case CH_F1: // F1
	send_func('P');
	break;
      case CH_F2: // F2
	send_func('Q');
	break;
      case CH_F3: // F3
	send_func('R');
	break;
      case CH_F4: // F4
	send_func('S');
	break;
      case CH_F5: // F5
	send_func('t');
	break;
      case CH_F6: // F6
	send_func('u');
	break;
      case CH_F7: // F7
	send_func('v');
	break;
      case CH_F8: // F8
	send_func('l');
	break;
	
      case 0xAB: // cmd + q
	break;
	
      case 0xAC: // cmd + d
	if(baud_index>0) --baud_index;
	ser_set();
	title();
	break;
	
      case 0xB8: // cmd + u
	if(baud_index<baud_idx_max) ++baud_index;
	ser_set();
	title();
	break;
	
#ifdef DEBUG
      case 0xBC: // cmd + c
	sprintf(STR_BUF,"                                        ");
	PutTitleAt(48);
	break;
#endif
	
      case 0x0D:
	ser_put_char(0x0D);
	if(priv_mode[0]&ANSI_LNM) {
	  ser_put_char(0x0A);
	}
	break;
	
      default:
	ser_put_char(c);
	break;
      }
    }
    
    if(send_get!=send_put) {
      ser_send();
    }
    //    if(c=='q') break;
  } // main loop
  
  Bitmap_Dis();
  
  return 0;
}
